#include "Arbitrator.h"

#include "ItemData.h"
#include "CellWidget.h"
#include "CellData.h"
#include "MenuWidget.h"
#include "MenuData.h"
#include "SettingsWidget.h"
#include "MenokuWidget.h"
#include "MenokuTray.h"
#include "IconFactory.h"
#include <QMessageBox>
#include <QInputDialog>
#include <QFile>
#include <QTextStream>

//These are global details about the root menu; Arbitrator needs them
//in order to update the menu file when changes are made
//declared in main.cpp
extern MenuData* root;
extern QString menuFile;

Arbitrator::Arbitrator() {
	//there is no selection
    m_cell = NULL;
    m_index = -1;

	//we start off NOT in edit mode
    m_edit = false;
}

//this makes connections to all nine cells in the main Menoku window
//(but not the clipboard cell!)
void Arbitrator::supplyCells(CellWidget** cells) {
    for (int i = 0; i < 9; i++) {
        connect(this,SIGNAL(setEdit(bool)),
                cells[i],SLOT(setEdit(bool)));
        connect(this,SIGNAL(invalidateCells()),
                cells[i],SLOT(invalidate()));
        connect(cells[i],SIGNAL(toggleVisible()),
                m_menoku,SLOT(toggleVisible()));
    }
}

//the one and only MenuWidget
void Arbitrator::supplyMenu(MenuWidget* menu) {
    m_menu = menu;
    connect(m_menu,SIGNAL(toggleVisible()),
            m_menoku,SLOT(toggleVisible()));
}

//the one and only SettingsWidget
void Arbitrator::supplySettings(SettingsWidget* settings) {
    m_settings = settings;
    connect(this,SIGNAL(invalidateCells()),
            m_settings->clipboardCell,SLOT(invalidate()));
	connect(settings,SIGNAL(updateHotKey(QString)),
	        this,SIGNAL(updateHotKey(QString)));
}

//The big form that holds everything
void Arbitrator::supplyMenoku(MenokuWidget* menoku) {
    m_menoku = menoku;
    connect(this,SIGNAL(setEdit(bool)),
            m_menoku,SLOT(setEdit(bool)));
}

//The tray icon
void Arbitrator::supplyTray(MenokuTray* tray) {
	connect(this,SIGNAL(updateHotKey(QString)),
	        tray,SLOT(updateHotKey(QString)));
	connect(tray,SIGNAL(toggleVisible()),
	        m_menoku,SLOT(toggleVisible()));
}

void Arbitrator::toggleEdit() {
	//switch modes!
    m_edit = !m_edit;
	
    //m_settings needs to know first. It updates the global settings
    //that the other widgets need to see.
    m_settings->setEdit(m_edit);
	
	//let the iconFactory take a look at the current settings. If they
	//mean that the icons need to be redrawn, tell all the cells that
	//they should update themselves.
    if (iconFactory->updateSettings()) {
        emit(invalidateCells());
    }
	
	//This contacts all the other widgets. The cells may redraw themselves here
	//even if the icon factory didn't invalidate them. However, they will not
	//redraw themselves twice.
    emit(setEdit(m_edit));
	
	//if we had selected an item, deselect it.
    if (m_cell) {
        m_cell->setSelection(-1);
        m_cell = NULL;
    }
	
	//if we're closing the edit pane
	if (!m_edit) {
		//Try to save the menu file.
		QFile menu(menuFile);
		if (menu.open(QIODevice::WriteOnly | QIODevice::Text)) {
			QTextStream(&menu) << root->toString();
		}
		else {
			//if it fails, report an error.
			QMessageBox::critical(0,
			                      "Could not save menu file",
								  "There was an error saving the menu file. Your changes HAVE NOT been saved.",
								  QMessageBox::Ok,QMessageBox::NoButton);
		}
	}
	
	//make sure that the menu has focus now so the user can signal it
	//with the keypad.
    m_menu->setFocus();
}

bool Arbitrator::edit() {
    return m_edit;
}

void Arbitrator::menuChanged() {
	//The menu has changed. Ask the menu widget for the new MenuData object.
    MenuData* menu = m_menu->m_data;
    for (int i = 0; i < 9; i++) {
		//go through every item in the clipboard cell and change its parent
		//to be the current menu. This makes dragging from the clipboard to
		//a cell in the MenuWidget trivial.
        ItemData* item = m_settings->clipboardCell->m_data->item(i);
        if (item && item->type() == ItemData::ItemMenu) {
            item->menu()->setParent(menu);
        }
    }
}

MenuData* Arbitrator::menu() {
	//The MenuWidget knows which menu is currently loaded.
    return m_menu->m_data;
}

void Arbitrator::requestActivation(CellWidget* cell,int index) {
	//if we're in edit mode, don't activate the item, but instead select
	//it so that it can be edited.
    if (m_edit) {
		//if they've clicked an already selected item (ie, clicked an item twice)
        if (m_cell == cell && m_index == index &&
			//and that item they clicked is a menu...
            cell->m_data->item(index)->type() == ItemData::ItemMenu) {
			//forget the selection and load that menu! This lets the user
			//traverse the menu hierarchy even in edit mode.
            clearSelection();
            cell->activate(index);
            return;
        }
		//if we had a selection before, the old cell needs to redraw itself
		//so that it no longer has the old selection marked.
        if (m_cell) {
            m_cell->setSelection(-1);
        }
		//now we make the new selection.
        cell->setSelection(index);
        m_cell = cell;
        m_index = index;
		//and prepare it to be edited.
        m_settings->setItem(cell->m_data->item(index));
    }
    else {
		//if we're not in edit mode, just activate the item!
        cell->activate(index);
    }
}

void Arbitrator::emptyActivation(CellWidget* cell,int index) {
	//If we're in edit mode and the user clicked on an empty space
    if (m_edit) {
		//clear the old selection
        if (m_cell) {
            m_cell->setSelection(-1);
        }
		//and make a brand new item in the space where the user clicked!
        cell->m_data->setItem(index,new ItemData("Unnamed","none"));
		//redraw the cell so that the new icon shows up
		//(I use loadCell to make sure internal variables regarding the
		//number of items is updated)
		cell->loadCell(cell->m_data);
		//and select that item!
        cell->setSelection(index);
        m_cell = cell;
        m_index = index;
        m_settings->setItem(cell->m_data->item(index));
    }
}

void Arbitrator::clickedCell(CellWidget* cell) {
	//if we're in edit mode...
    if (m_edit) {
		//clear the selection
        clearSelection();
        bool ok;
		//and ask the user to rename the cell he right clicked on.
        QString newName = QInputDialog::getText(0,
                                                "Rename cell",
                                                "Enter a new name for this cell:",
                                                QLineEdit::Normal,
                                                cell->m_data->name(),
                                                &ok);
		//if the user gave a valid name, apply it.
        if (ok && !newName.isEmpty()) {
            cell->m_data->setName(newName);
            cell->drawBody();
            cell->update();
        }
    }
}

void Arbitrator::applyItem() {
	//the user just applied changes to an item, so it must be redrawn.
    //(force the cell to be loaded from scratch so that various its
	//internal state can be updated)
	m_cell->loadCell(m_cell->m_data);
	//and then deselect.
    clearSelection();
}

void Arbitrator::cancelItem() {
    clearSelection();
}

void Arbitrator::deleteItem() {
	//remove the item from the cell
    m_cell->m_data->clearItem(m_index);
	//then redraw the cell to reflect this change
	m_cell->loadCell(m_cell->m_data);
	//and deselect.
    clearSelection();
}

void Arbitrator::clearSelection() {
	//tell the SettingsWidget to forget what it's doing
    m_settings->clearItem();
	//if we had something selected,
    if (m_cell) {
		//deselect it and forget about it.
        m_cell->setSelection(-1);
        m_cell = NULL;
	    m_index = -1;
    }
	//send focus back to the MenuWidget so the user can select items
	//with the keyboard again.
    m_menu->setFocus();
}

