#include "CellData.h"
#include "ItemData.h"

CellData::CellData(const QString &name)
{
	m_name = name;
	m_count = 0;
	//unset items must be NULL so I know they're not set.
	for (int i = 0; i < 9; i++) {
		m_items[i] = NULL;
	}
	m_item = NULL;
	m_item_index = -1;
}

CellData::CellData(CellData* cell) {
	//copy from another cell instead of using default values
    m_name = cell->m_name;
    m_count = cell->m_count;
    for (int i = 0; i < 9; i++) {
        if (cell->m_items[i] == NULL) {
            m_items[i] = NULL;
        }
        else {
            m_items[i] = new ItemData(cell->m_items[i]);
        }
    }
    m_item_index = cell->m_item_index;
    m_item = m_items[m_item_index];
}

void CellData::copy(CellData* cell) {
	//This is identical to the copy constructor, except...
    m_name = cell->m_name;
    m_count = cell->m_count;
    for (int i = 0; i < 9; i++) {
		//I delete any items that have already been set! Very important.
        if (m_items[i]) {
            delete m_items[i];
        }
        if (cell->m_items[i] == NULL) {
            m_items[i] = NULL;
        }
        else {
            m_items[i] = new ItemData(cell->m_items[i]);
        }
    }
    m_item_index = cell->m_item_index;
    m_item = m_items[m_item_index];
}

CellData::~CellData()
{
	//so the tree can recursively delete itself:
    for (int i = 0; i < 9; i++) {
        if (m_items[i])
            delete m_items[i];
    }
}

QString const CellData::name() {
	return m_name;
}

void CellData::setName(const QString &name) {
    m_name = name;
}

int const CellData::count() {
	return m_count;
}
	
ItemData* CellData::item(int index) {
	return m_items[index];
}

ItemData* CellData::item() {
	return m_item;
}

int CellData::itemIndex() {
    return m_item_index;
}

void CellData::setItem(int index, ItemData* item) {
	//if this spot was already filled, delete the item that filled it
    if (m_items[index]) {
        delete m_items[index];
    }
    else {
		//otherwise we have one more item in this cell than before.
    	++m_count;
    }

	//update the sample item.
	m_item = item;
	m_item_index = index;
	m_items[index] = item;
}

void CellData::clearItem(int index) {
	//if the item needs to be cleared,
    if (m_items[index]) {
		//remember the old pointer.
        ItemData* old = m_items[index];
		//delete the item.
        delete m_items[index];
		//and mark this item as empty.
        m_items[index] = NULL;
        --m_count;
		//if we just cleared the sample item, we need to
		//update the sample item so it points to SOMETHING.
        if (m_item == old) {
            m_item = NULL;
            m_item_index = -1;
			//if there's another item to point to,
			if (m_count > 0) {
				//find it and make it our sample item.
	            for (int i = 0; i < 9; i++) {
    	            if (m_items[i]) {
        	            m_item = m_items[i];
            	        m_item_index = i;
						break;
					}
                }
            }
        }
    }
}

void CellData::clear() {
	//clear all the items.
    for (int i = 0; i < 9; i++) {
        if (m_items[i]) {
            delete m_items[i];
            m_items[i] = 0;
        }
    }
	//we are now empty
    m_count = 0;
	//so there is no sample item.
    m_item = NULL;
    m_item_index = -1;
	//forget the name, too.
    m_name = "Unnamed";
}

QString CellData::toString(int pos, const QString &indent) {
    QString e_name = m_name;
	//Since the name will appear within double quotes, the only invalid character
	//the user could enter would be a double quote char. So, we must escape these.
    e_name.replace(QChar('"'),"&quot;");
	//build up the cell tag...
	QString result = indent + "<cell pos=\"" + QString::number(pos) + "\" name=\"" + e_name + "\">\n";
	//now go through the items in the order they appear on the numpad
	//(7, 8, 9, 4, 5, 6, 1, 2, 3)
	for (int start = 6; start >= 0; start -= 3) {
		for (int i = start; i < start+3; i++) {
			if (m_items[i]) {
				//print each cell. Note that we tell the cell which position
				//it's in so it doesn't have to remember. This makes drag
				//and drop simpler.
				result += m_items[i]->toString(i+1,indent+"\t");
			}
		}
	}
	//close the cell tag.
	result += indent + "</cell>\n";
	return result;
}
