#include "CellWidget.h"

#include "CellData.h"
#include "ItemData.h"
#include "IconFactory.h"
#include "Arbitrator.h"
#include "DefaultSettings.h"
#include <QApplication>
#include <QMessageBox>
#include <QPainter>
#include <QKeyEvent>
#include <QMouseEvent>
#include <QSettings>
#include <QProcess>
#include <QMenu>
#include <qdrawutil.h>

//a simple macro to save a bit of typing. Finds the largest multiple of RANGE
//that X is greater than or equal to.
#define NEAREST(X,RANGE) (((X)/(RANGE))*(RANGE))

using namespace Qt;

//constants for naming new Mime types for dragging.
const char* DRAG_ITEM = "menoku item drag";
const char* DRAG_CELL = "menoku cell drag";

//shared information about dragging/dropping
ItemData* CellWidget::m_swapped_item = NULL;
CellData* CellWidget::m_swapped_cell = NULL;
int CellWidget::m_drag_index = -1;


CellWidget::CellWidget(QWidget* parent)
 : QWidget(parent) {
    m_data = NULL;
    m_empty = true;
    m_one_item = false;
    m_dark = false;
    m_invalid = false;

    m_title_h = 0;
	//these settings will be loaded from QSettings at the beginning of the
	//program (The edit state will be turned to OFF and setEdit(bool) will
	//get called)
    m_speed_single_item = false;
    m_zoom_single_item = false;
    m_hide_empty_cell = false;

    //turn off double buffering
   	setAttribute(WA_NoSystemBackground,true);
   	setAcceptDrops(true);
	//So we can draw highlights under the mouse:
   	setMouseTracking(true);
}

void CellWidget::drawBody() {
	//if we're drawing the body, we're no longer invalid
    m_invalid = false;;
	//rebuild the pixmap from scratch and make sure it's the right size
	//this method is called on a resize event, too.
    m_body = QPixmap(size());
    m_body.fill(this,0,0);

    QPainter p(&m_body);

    //some of this computation could probably be stored and used elsewhere.
	//compute the font for the cell header
	QFont font_normal = p.font();
	QFont font_big = p.font();
	font_big.setBold(true);
	font_big.setPointSize(font_normal.pointSize()+4);

	p.setFont(font_big);

	//compute the size of the header and, considering that, all the other
	//important dimensions
	QRect bound = p.boundingRect(0,2,width(),0,AlignHCenter,m_data->name());
	m_title_h = bound.height() + 2;
	m_block_h = (height() - m_title_h) / 3;
	m_block_w = width() / 3;

    //If this cell is empty and we don't want to draw empty cells, we can stop
	//here now that the dimensions have been calculated.
    if (m_empty && m_hide_empty_cell) return;
	//shade the background for the cell's title.
	p.fillRect(QRect(0,0,width(),m_title_h),palette().color(QPalette::Mid));
	//draw a bevel around the whole cell
	qDrawWinPanel(&p,0,0,width(),height(),palette());
	//draw the name of the cell at the top.
	p.drawText(bound,AlignHCenter,m_data->name());

	//Now that the body's drawn, revert back to the normal font for drawing
	//the titles under the icons
    p.setFont(font_normal);

	//if there's only one item and we're supposed to draw single items big...
    if (m_one_item && m_speed_single_item) {
		//fetch the icon,
        ItemData* item = m_data->item();
        QPixmap icon = iconFactory->iconFor(item);
		//scale it up
        if (m_zoom_single_item) {
            icon = icon.scaled(icon.width()*3,icon.height()*3);
        }
		//and draw it using the entire space.
        drawIcon(p,0,m_title_h,width(),height() - m_title_h,icon,item->name(),item->type() == ItemData::ItemMenu);
    }
    else {	
		//otherwise, draw all the items separately.
		//for each of the nine spots,
        for (int i = 0; i < 9; i++) {
			//if there's an item there,
            ItemData* item = m_data->item(i);
            if (item) {
				//calculate the region it should ocupy,
                int x = (i % 3) * m_block_w;
                int y = (2 - (i/3)) * m_block_h + m_title_h;
				//lookup the icon
                QPixmap icon = iconFactory->iconFor(item);
				//and draw it.
                drawIcon(p,x,y,m_block_w,m_block_h,icon,item->name(),item->type() == ItemData::ItemMenu);
            }
        }
    }
}

void CellWidget::setDark(bool dark) {
	//remember that I'm set dark
    m_dark = dark;
	//unhighlight myself
	m_highlight.setRect(0,0,0,0);
	//and draw these changes
    update();
}

void CellWidget::invalidate() {
    m_invalid = true;
}

void CellWidget::setSelection(int index) {
	//an index of -1 means deselect.
    if (index != -1) {
		//figure out the region that should be marked
		//(could I combine this code with the highlighting code?)
        int x = (index % 3) * m_block_w + 1;
        int y = (2 - (index / 3)) * m_block_h + m_title_h + 1;
		//and set the rectangle to mark selections with.
        m_selection.setRect(x,y,m_block_w - 2,m_block_h - 2);
    }
    else {
		//if we were meant to deselect, make the rectangle empty.
        m_selection.setRect(0,0,0,0);
    }
	//either way, draw the change.
    update();
}

int CellWidget::heightForWidth(int w) {
    return w + m_title_h;
}

void CellWidget::resizeEvent(QResizeEvent*) {
    drawBody();
}

void CellWidget::leaveEvent( QEvent * )
{
	//if the mouse was over one of the items, we must unhighlight that area.
	if (m_highlight.isValid()) {
    	m_highlight.setRect(0,0,0,0);
    	update();
    }
}

int CellWidget::indexAt(const QPoint &point) {
	//compute the x and y coordinates (in terms of the dimensions of one item)
    int x = point.x() / m_block_w;
    int y = (point.y() - m_title_h) / m_block_h;
	//if they are in range of the items and not clicking on the title,
    if (x >= 0 && x < 3 && y >= 0 && y < 3 && point.y() > m_title_h) {
		//convert from x/y into indices.
        return (2-y) * 3 + x;
    }
    else {
		//otherwise, they're out of range. -1.
        return -1;
    }
}

ItemData* CellWidget::itemAt(const QPoint &point) {
	//if there's only one item and we're letting the item take over the rest of
	//the cell
    if (m_one_item && m_speed_single_item) {
		//then it doesn't matter where they clicked, they want that one item.
        return m_data->item();
    }
    else {
		//otherwise, figure out what index it was.
        int index = indexAt(point);
		//and properly return either the item there or NULL.
        if (index != -1 && m_data->item(index)) {
            return m_data->item(index);
        }
        else {
            return NULL;
        }
    }
}

void CellWidget::mousePressEvent(QMouseEvent* e) {
	//remember where the drag started to see if it really is a drag
    m_drag_start = e->pos();
}

void CellWidget::mouseReleaseEvent(QMouseEvent* e) {
	//if they left clicked on the cell,
    if (e->button() == LeftButton) {
		//if there is one item and we're letting that item take over the cell,
        if (m_one_item && m_speed_single_item) {
			//then they must have meant to activate that item.
			//(note, this will never happen in edit mode because
			//m_speed_single_item will be false)
            boss->requestActivation(this,m_data->itemIndex());
        }
        else {
			//otherwise, figure out where they clicked.
            int index = indexAt(e->pos());
            if (index == -1) {
				//if they clicked on the cell but not one of its items, report that.
                boss->clickedCell(this);
            }
            else {
				//if they picked an item, though, go with it!
                choose(index);
            }
        }
    }
    else {
		//any other mouse button means go up to the previous menu.
        emit up();
    }
}

void CellWidget::mouseMoveEvent( QMouseEvent * e )
{
	//if this cell isn't being updated because it's empty or dark, don't worry.
    if ((m_empty && m_hide_empty_cell) || m_dark) return;
	//if the mouse is being dragged
    if (e->buttons() & LeftButton) {
		//if we're in edit mode and we've passed the critical distance to start
		//a drag event, then we are dragging an icon.
	    if (boss->edit() && (e->pos() - m_drag_start).manhattanLength() > QApplication::startDragDistance()) {
			//figure out which item they want to drag.
            int index = indexAt(m_drag_start);
			//if they're dragging an item
            if (index != -1) {
                ItemData* item;
				//we make a duplicate of this item, just in case they want to copy it.
				//it's important that we don't have two cells pointing at the same
				//ItemData object if we don't want memory errors all over the place.
                if (m_one_item && m_speed_single_item) {
                    item = new ItemData(m_data->item());
                }
                else {
                    if (!m_data->item(index)) return;
                    item = new ItemData(m_data->item(index));
                }
				//create a new drag object
                QDrag *drag = new QDrag(this);
				//and associate with it the newly copied item to drag.
                QMimeData *mimeData = new QMimeData;
				//the data associated with it is a pointer to the item to
				//drag, masquerading as an int.
                mimeData->setData(DRAG_ITEM,QByteArray::number((int) item));
				//remember which item we're dragging.
                m_drag_index = index;

				//set up the details for this drag item.
                drag->setMimeData(mimeData);
                QPixmap icon = iconFactory->iconFor(item).scaled(20,20);
                drag->setPixmap(icon);
				//it can be copied, moved, or linked (link == swap)
                DropAction dropAction = drag->start(CopyAction | MoveAction | LinkAction);
				//at this point, the drag has been completed and the item dropped!
				//what we do depends on which action was used.
                switch (dropAction) {
					//if it was a move, we must now empty the cell the drag
					//started from.
                    case MoveAction:
                        m_data->clearItem(index);
						//reload this cell, counting how many items are in it and redrawing it.
						loadCell(m_data);
                        break;
					//if it was a swap we must now put the displaced item into
					//the empty slot.
                    case LinkAction:
                        if (m_swapped_item) {
                            m_data->setItem(index,m_swapped_item);
							//if we swapped an item for an item, icons changed so redraw.
	                        drawBody();
    	                    update();
                        }
                        else {
							//if the icon was dragged to an empty location, just
							//clear the spot where it was dragged from.
                            m_data->clearItem(index);
							//if we swapped an item for nothing, then maybe m_empty
							//should be set now. Reload the cell.
							loadCell(m_data);
                        }
                        break;
                    case CopyAction:
                    default:
                        //if the action was a copy or nothing happened, nothing
						//needs to take place at this end.
                        return;
                }
            }
			//otherwise, they're dragging the whole cell!
            else {
				//the process is pretty similar to above, though a little simpler.
                QDrag *drag = new QDrag(this);
                QMimeData *mimeData = new QMimeData;
                mimeData->setData(DRAG_CELL,QByteArray::number((int) m_data));

                drag->setMimeData(mimeData);
                DropAction dropAction = drag->start(CopyAction | MoveAction | LinkAction);
                switch (dropAction) {
                    case MoveAction:
                        m_data->clear();
                        drawBody();
                        update();
                        break;
                    case LinkAction:
                        m_data->copy(m_swapped_cell);
                        delete m_swapped_cell;
                        loadCell(m_data);
                        break;
                    case CopyAction:
                    default:
                        return;
                }
            }
	    }
	}
	//if the user is just moving the mouse and NOT dragging with the left button,
    else {
		//then update the highlight to follow the mouse!
	    if (!m_highlight.contains(e->pos())) {
	        if (indexAt(e->pos()) != -1) {
                if (m_one_item && m_speed_single_item) {
                    m_highlight.setRect(0,m_title_h,width(),height() - m_title_h);
                }
                else {
                    m_highlight.setRect(NEAREST(e->x(),m_block_w),
                                        NEAREST(e->y() - m_title_h,m_block_h) + m_title_h,
                                        m_block_w,
                                        m_block_h);
                }
            }
			//select nothing if the mouse isn't over some item.
            else {
                m_highlight.setRect(0,0,0,0);
            }
            update();
        }
	}
}

void CellWidget::dragEnterEvent(QDragEnterEvent* e) {
	//This seems to be needed in Qt4.1  :)
	e->accept();
}

void CellWidget::dragMoveEvent(QDragMoveEvent* e) {
	//normally, we don't accept drags.
    e->ignore();
	//however, if it's one of the recognized types, we might.
    if (e->provides(DRAG_ITEM)) {
        int index = indexAt(e->pos());
		//if they're dragging an item over a valid item spot
		//AND that spot isn't where the drag began
        if (index != -1 &&
            (e->source() != this || m_drag_index != index)) {
			//then we'll accept this drop.
            e->acceptProposedAction();
        }
    }
    else if (e->provides(DRAG_CELL) && e->source() != this) {
		//you can drop a cell wherever you please.
        e->acceptProposedAction();
    }
}

void CellWidget::dropEvent(QDropEvent* e) {
	//if there was no modifier pressed...
	if ((e->keyboardModifiers() & (ShiftModifier | ControlModifier | AltModifier)) == 0) {
		//create a popup menu asking the user which action to perform.
		QMenu popup;
		//the menu itself gives hints about the key modifiers you can use to avoid
		//this menu.
		QAction* copy = popup.addAction("Copy (Ctrl)");
		QAction* move = popup.addAction("Move (Shift)");
		QAction* swap = popup.addAction("Swap (Alt)");
		QAction* result = popup.exec(mapToGlobal(e->pos()),copy);
		//set the action based on what the user selected.
		if (result == copy)
			e->setDropAction(CopyAction);
		else if (result == move)
			e->setDropAction(MoveAction);
		else if (result == swap)
			e->setDropAction(LinkAction);
		else {
			e->ignore();
			return;
		}
	}
	else {
		//otherwise, they used a modifier to tell us which action to take.
		//Do whatever it is they asked for.
		e->setDropAction(e->proposedAction());
	}

	//if they've just dropped an item,
    if (e->provides(DRAG_ITEM)) {
		//make sure they're dropping at an appropriate spot
		//(this may be redundant)
        int index = indexAt(e->pos());
        if (index == -1) {
            e->ignore();
            return;
        }
    
		//if they're swapping, we need to swap out the old item.
        if (e->dropAction() == LinkAction) {
            if (m_data->item(index)) {
				//note that we make a copy. This is to avoid pointer hell and
				//multiple references to the same item.
                m_swapped_item = new ItemData(m_data->item(index));
            }
            else {
                m_swapped_item = NULL;
            }
        }
    
		//recover the ItemData* that we passed...
        ItemData* item = (ItemData*) e->encodedData(DRAG_ITEM).toInt();
		//and put that item in place. Note, this will not screw up any swapped
		//items because we made a copy  ;)
        m_data->setItem(index,item);
    
        //if we just dragged from ourself to ourself, then the drop end of things
		//is going to try to redraw this cell anyway. That is, unless it was a
		//copy, then it won't.
        if ((e->source() != this) ||
            (e->dropAction() == CopyAction)) {
			//so, if the drop isn't going to redraw this cell anyway, we must
			//explicitely redraw it here.
			//Since adding a new item might change variables like m_empty and such,
			//reset all those variables.
            loadCell(m_data);
        }
		e->accept();
    }
	//otherwise, they must be dropping a cell.
    else {
		//again, much like the above code, only simpler.
        if (e->dropAction() == LinkAction) {
            //must make a duplicate so changing m_data won't affect the swap.
            m_swapped_cell = new CellData(m_data);
        }

        CellData* cell = (CellData*) e->encodedData(DRAG_CELL).toInt();
        if (cell == m_data) {
            e->ignore();
            return;
        }
        m_data->copy(cell);
        loadCell(m_data);
 
		e->accept();
    }
}

void CellWidget::paintEvent(QPaintEvent *e) {
	//if we need to redraw everything from scratch, do so!
    if (m_invalid) drawBody();

    QPainter p(this);

	//paint the the invalidated region from the buffer
    p.drawPixmap(e->rect(),m_body,e->rect());

	//if we're darkened, darkent the cell.
    if (m_dark) {
		//since this is usually most of the screen and that takes far too long
		//to draw with transparency, I'll use dithering instead.
        p.fillRect(rect(),QBrush(palette().color(QPalette::Shadow),Dense4Pattern));//QColor(0,0,0,80));
    }
    else if (m_highlight.isValid()) {
		//since this is usually one small area and it looks much better with
		//transparency, I'll use transparency here.
        p.fillRect(m_highlight,QColor(255,255,255,80));
    }
    
	//if there is a selected item, mark it as selected.
    if (m_selection.isValid()) {
        p.setBrush(NoBrush);
        p.setPen(QPen(QBrush(black),2));
        p.drawRect(m_selection);
    }
}

void CellWidget::drawIcon(QPainter &painter,
			              const int x, const int y, const int w, const int h,
                          const QPixmap &pixmap, const QString &label,
						  bool menu)
{
	//the dimensions of the pixmap
 	int pixw = pixmap.width();
	int pixh = pixmap.height();
	//if the icon itself will fit in the space we're given, draw it.
	if (h > pixh && w > pixw) {
		//find out how much space the icon label will take up.
		QRect bound = painter.boundingRect(0,0,0,0,AlignHCenter|AlignTop,label);
		int texth = bound.height();
		//the position of the pixmap
		int offx = (w - pixw) / 2;
		int offy = (h - pixh) / 2;

		//if there's no room to fit the pixmap AND the text, just draw the pixmap
		if ((h-pixh) < texth) {
			painter.drawPixmap(x+offx,y+offx,pixmap);
		}
		else {
			//otherwise, make room for the text and recenter things to
			//accomodate it.
			offy = (h - pixh - texth) / 2;
			painter.drawPixmap(x+offx,y+offy,pixmap);
		
			//if the label is too long, cut off letters from the end until it fits.
			QString new_label = label;
			int len = label.length();
			//there might be a better way to do this, but this method is more
			//than fast enough.
			while (bound.width() > w) {
				new_label = label.left(--len) + "...";
				bound = painter.boundingRect(0,0,0,0,AlignHCenter|AlignTop,new_label);
			}
			//draw the text.
			painter.drawText(x,y+offy+pixh,w,texth,AlignHCenter|AlignTop,new_label);
		}
		//if we're drawing an item that will launch a submenu, mark it as such.
		if (menu) {
			QPixmap marker(":menumarker.png");
			int mx = x+offx+pixw;
			int my = y+offy+pixh - marker.height();
			painter.drawPixmap(mx,my,marker);
		}
	}
}

void CellWidget::loadCell(CellData* cell) {
	//switch over to this new cell. Do not delete the old cell, the entire menu
	//must be kept in tact.
    m_data = cell;
    m_empty = (m_data->count() == 0);
    m_one_item = (m_data->count() == 1);

    drawBody();
    update();
}

void CellWidget::choose() {
	//if there's only one item and we want to let it take over, then we don't need
	//a second keypress to know which item they want.
    if (m_one_item && m_speed_single_item) {
        boss->requestActivation(this,m_data->itemIndex());
    }
	//otherwise, go ahead and focus in on this cell and let the user pick one of
	//its items. That is, unless it's hidden because it's empty.
    else if (!m_empty || !m_hide_empty_cell) {
        emit(select(this));
    }
}

void CellWidget::choose(int index) {
	//if there is an item at that index
    if (m_data->item(index)) {
		//ask the boss what to do about it.
        boss->requestActivation(this,index);
    }
    else {
		//if it's an empty spot, tell the boss what happened.
        boss->emptyActivation(this,index);
    }
	//in either case, an item was picked. Let the menu take focus back now.
    emit(deselect());
}

void CellWidget::activate(int index) {
    ItemData* item = m_data->item(index);
	//if the item to activate was an application
    if (item->type() == ItemData::ItemExec) {
		//try to start the process! In my experience, this never returns false,
		//which is a shame and frustrating. Still, it should return false if
		//something goes wrong, so I have a handler for that.
		//Note, you can't call startDetached with a null string, so I also make 
		//sure that there is some application listed.
		if (item->exec().isEmpty()) return;
        if (QProcess::startDetached(item->exec())) {
			emit(toggleVisible());
		}
		else {
			QMessageBox::critical(0,
			                      "Failed to execute program",
								  "There was an error launching: " + item->exec(),
								  QMessageBox::Ok,QMessageBox::NoButton);
		}
    }
	//otherwise, it's a menu! Load that menu now.
    else {
        emit(loadMenu(item->menu()));
    }
}

void CellWidget::setEdit(bool edit) {
	//remember the old values for all the settings.
	//If they don't change, we don't need to redraw anything.
    bool m_old_speed_single_item = m_speed_single_item;
    bool m_old_zoom_single_item = m_zoom_single_item;
    bool m_old_hide_empty_cell = m_hide_empty_cell;
	//in edit mode, we switch to the simplest settings. This makes drag and drop
	//sooooo much easier.
    if (edit) {
        m_speed_single_item = false;
        m_zoom_single_item = false;
        m_hide_empty_cell = false;
    }
    else {
		//when we go out of edit mode, we must reload all the settings just in
		//case they changed.
        QSettings settings;
        settings.beginGroup(CELL_OPTIONS);
        m_speed_single_item = settings.value(SPEED_SINGLE_ITEM,SPEED_SINGLE_ITEM_DEFAULT).toBool();
        m_zoom_single_item = settings.value(ZOOM_SINGLE_ITEM,ZOOM_SINGLE_ITEM_DEFAULT).toBool();
        m_hide_empty_cell = settings.value(HIDE_EMPTY_CELL,HIDE_EMPTY_CELL_DEFAULT).toBool();
    }

	//don't bother with all of this if we're hidden. This avoids some redundant
	//calculations.
    if (isVisible()) {
		//if options changed and those options affect how this cell is drawn,
		//then invalidate ourselves. Note, it's quite possible that some icon
		//settings changed and this cell has already been invalidated. That's
		//why the new value is ORed with the old one.
        m_invalid = m_invalid ||
                    (m_one_item &&
                     ((m_old_speed_single_item != m_speed_single_item) ||
                      (m_old_zoom_single_item != m_zoom_single_item)))
                    ||
                    (m_empty &&
                     ((m_old_hide_empty_cell != m_hide_empty_cell)));
		//this will call drawBody() if necessary.
        update();
    }
}

void CellWidget::keyPressEvent( QKeyEvent * e )
{
	int index = 0;
	switch (e->key()) {
        case Key_Escape:
			//hide the form if escape was pressed.
            emit( toggleVisible() );
            //go ahead and deselect while we're at it...
		case Key_0:
		case Key_Insert:
			//0 goes back a level.
			emit(deselect());
			return;
		//all the number-pad keys choose their specic items.
		//Note that all of these have their numlock-on and numlock-off aliases
		//here. That makes sure the program will work with or without numlock.
		//it also means that the other number keys and such will also work.
		case Key_1:
		case Key_End:
			index = 0;
			break;
		case Key_2:
		case Key_Down:
			index = 1;
			break;
		case Key_3:
		case Key_PageDown:
			index = 2;
			break;
		case Key_4:
		case Key_Left:
			index = 3;
			break;
		case Key_5:
		case Key_Clear:
			index = 4;
			break;
		case Key_6:
		case Key_Right:
			index = 5;
			break;
		case Key_7:
		case Key_Home:
			index = 6;
			break;
		case Key_8:
		case Key_Up:
			index = 7;
			break;
		case Key_9:
		case Key_PageUp:
			index = 8;
			break;
		//any other key should be ignored.
		default:
            return;
	}
	//the user picked a cell! Try to activate it.
	choose(index);
}
