#ifndef CELL_WIDGET_H
#define CELL_WIDGET_H

#include <QWidget>
#include <QPixmap>

class ItemData;
class CellData;
class MenuData;

//This is the Widget that graphically represents a cell (a group of nine icons)
//Since this handles drag and drop, is responsible for drawing icons, and
//is frequently adjusted for the user, it is by far the most complicated class.
//Note that there is no ItemWidget. CellWidget is responsible for drawing all
//of its items because having 81 extra QWidgets is a little hard on system
//resources  ;)
class CellWidget : public QWidget {
	//Let the arbitrator pick into the guts of this class as much as it wants.
	//This class is complicated enough as it is.
	friend class Arbitrator;
	Q_OBJECT
	public:
    	CellWidget(QWidget *parent = 0);
		
		//if a cell is set dark, it will not highlight itself when the mouse
		//passes by and it will appear darkened. This is used to highlight
		//which cells AREN'T currently selected by the user.
   		void setDark(bool dark);
		//mark one of the items in this cell as being edited.
   		void setSelection(int index);
		
		//how high a cell should be, optimally, for a given width.
   		int heightForWidth(int w);

		//Set the CellData object associated with this CellWidget
   		void loadCell(CellData* cell);

	signals:
		//a new menu should be loaded
    	void loadMenu(MenuData* menu);
		//the user has requested to go up a level in the menu system
    	void up();
		//this cell has been chosen. Set the other cells to be dark
		//and let THIS cell capture keystrokes.
    	void select(CellWidget* cell);
		//undarken all the cells and let the MenuWidget handle keystrokes.
    	void deselect();
		//the visibility state of the Menoku window should change
		//(since this will only happen when the window is visible, that
		//means hide the Menoku window)
    	void toggleVisible();

	public slots:
		//This cell was selected by the user's typing
    	void choose();
		//the indexth item in this cell was selected by the user's typing
    	void choose(int index);
		//activate the indexth item. If it is a menu, load it. If it is an
		//application, start it!
    	void activate(int index);
		//report when the application enters or exits edit mode.
    	void setEdit(bool edit);
		//force this widget to redraw itself when it gets the chance to.
		//This method exists to avoid redundant redrawing when several changes
		//need to happen at roughly the same time.
    	void invalidate();

	protected:
		//the whole widget must be redrawn on a resize event.
 		void resizeEvent(QResizeEvent*);

		//draw all the icons, labels, etc. for this Cell onto its buffer pixmap.
		//when the cell is updated, the buffer pixmap is used. I do this instead
		//of using Qt4's built in buffering because I often need to draw things
		//on TOP of the buffer pixmap (highlights, shadow) without needing to
		//redraw the icons themselves.
    	void drawBody();

		//what index item would be drawn beneath the given point?
		//returns -1 if no item would be drawn at that point.
		//this does not guarantee that there is an item at the returned index
    	int indexAt(const QPoint &point);
		//like above, except it actually looks up the item that would be under
		//the point. Returns null if there is no real item drawn at the given
		//point.
    	ItemData* itemAt(const QPoint &point);

		//mouse events for handling highlights and drag and drop
		void leaveEvent(QEvent* e);
		void mouseMoveEvent(QMouseEvent* e);
		void mousePressEvent(QMouseEvent* e);
		void mouseReleaseEvent(QMouseEvent* e);
		void dragEnterEvent(QDragEnterEvent* e);
		void dragMoveEvent(QDragMoveEvent* e);
		void dropEvent(QDropEvent* e);

		//paint this widget. In most cases, this just means copying from
		//the buffer pixmap and maybe drawing some extra stuff on top.
    	void paintEvent(QPaintEvent* e);
		//draw one icon with some labeling text within the specified rectangle
		//on the specified painter. If the item is a menu, draw the menumarker
		//pixmap next to the icon to show this.
    	void drawIcon(QPainter &painter,
		        	  const int x, const int y, const int w, const int h,
                	  const QPixmap &pixmap, const QString &label, bool menu);

		//handle a key press! 1-9 on the keypad will try to activate an item,
		//0 will cancel the current selection, and escape will hide the window.
    	void keyPressEvent(QKeyEvent *e);

		//when dragging items/cells you have the option to swap the item you're
		//dragging for the one in the position you're dropping to. These let
		//different cell objects share the displaced items so they can be swapped
    	static ItemData* m_swapped_item;
    	static CellData* m_swapped_cell;
		//the index of the item that is being dragged (used to tell if you are
		//dropping to the same spot you dragged from)
    	static int m_drag_index;

		//true iff the cell this is representing has no items in it.
    	bool m_empty;
		//true iff the cell this is representing has one item in it.
    	bool m_one_item;

		//drawing options.
		//should selecting a cell with one item activate that one item without
		//the user having to select it?
    	bool m_speed_single_item;
		//if there is only one item, should it be scaled to fill the given space?
    	bool m_zoom_single_item;
		//if there are no items here, should I bother drawing anything?
    	bool m_hide_empty_cell;

		//is this cell currently set dark?
    	bool m_dark;
		//does drawBody() need to be called?
    	bool m_invalid;
		//the height of the title of this cell
		int m_title_h;
		//the width of an item
		int m_block_w;
		//the height of an item
		int m_block_h;
		//the extent of the space alloted to the item beneath the mouse.
		//(drawn lighter than the rest of the widget)
		QRect m_highlight;
		//the rectangle surrounding an item selected for editing
		QRect m_selection;
		//the point where the mouse first went down. This is to detect whether
		//the user is actually intending to drag or just accidentally moving
		//the mouse while clicking.
		QPoint m_drag_start;

		//the buffer pixmap containing the title, border, and icons of this cell
		QPixmap m_body;

		//the actual data this cell represents.
		CellData* m_data;
};

#endif //CELL_WIDGET_H
