#include "FileParser.h"
#include <QXmlSimpleReader>
#include <QXmlInputSource>
#include <QFile>
#include "ItemData.h"
#include "CellData.h"
#include "MenuData.h"

MenuData* FileParser::parse(const QString & fileName) {
	m_state = StateNone;
	m_root = NULL;
	m_menu = NULL;
	m_cell = NULL;

	//balk if the file isn't available.
	QFile file(fileName);
	if (!file.exists()) {
	    //can't use setError here because m_locator is not yet defined.
        m_error = "File name not found: " + fileName;
        return NULL;
	}

	//make a new reader, set it up to use this class for callbacks.
	QXmlSimpleReader reader;
	reader.setContentHandler(this);
	reader.setErrorHandler(this);
	QXmlInputSource input(&file);

	//parse the input file.
	if (reader.parse(input)) {
		//return the root menu if it worked
		return m_root;
	}
	else {
		//otherwise, delete the root menu (and everything else because of the
		//recursive delete) so we don't lose all that memory!
		if (m_root)
            delete m_root;
		//return NULL to show that we failed.
		return NULL;
	}
}

bool FileParser::startElement(const QString &, const QString &,
                          	  const QString &qName, const QXmlAttributes &attributes) {
	switch (m_state) {
		//if we just started parsing
		case StateNone:
			//only a rootmenu tag is valid.
			//at some point, I will want to check the version flag here, but for
			//now there is only one version and it makes no difference.
			if (qName == "rootmenu") {
				m_menu = new MenuData("ROOT",NULL);
				m_root = m_menu;
				m_state = StateMenu;
				return true;
			}
			else {
				setError("Menu file must begin with a <rootmenu> tag");
				return false;
			}
		//if we're parsing a cell
		case StateCell:
			//then that cell must contain only menu and exec tags.
			if ((qName == "menu") || (qName == "exec")) {
				//make sure that there is a position given and that it is valid.
				int pos = attributes.value("pos").toInt();
				if (pos < 1 || pos > 9) {
					setError("Invalid position given in <" + qName + "> tag. Please specify a number from 1 to 9.");
					return false;
				}
				//convert from 1-9 to 0-8
				--pos;
				//get the name,
				QString name = attributes.value("name");
				if (name.isEmpty()) {
					setError("No name specified in <" + qName + "> tag.");
					return false;
				}
				//get the icon, but don't panic if there isn't one specified.
				QString icon = attributes.value("icon");
				//if the spot is already full, complain.
				if (m_cell->item(pos)) {
					setError("Cannot add item \"" + name + "\" to already filled item position " + QString::number(pos) + ".");
					return false;
				}
				//now that we know we have a valid menu or exec tag, build the
				//proper ItemData object to fill the spot.
				if (qName == "menu") {
					//if this is a menu, then the next step is parsing that menu.
					m_menu = new MenuData(name,m_menu);
					m_item = new ItemData(name,icon);
					m_item->setMenu(m_menu);
					m_cell->setItem(pos,m_item);
					m_state = StateMenu;
				}
				else {
					m_item = new ItemData(name,icon);
					m_cell->setItem(pos,m_item);
					m_state = StateExec;
				}
				return true;
			}
			else {
				setError("Only <menu> and <exec> tags are allowed within a <cell> tag. Found <" + qName + "> tag instead.");
				return false;
			}
		//if we're in a menu
		case StateMenu:
			//we can only have other cells.
			//this is similar to the above code.
			if (qName == "cell") {
				int pos = attributes.value("pos").toInt();
				if (pos < 1 || pos > 9) {
					setError("Invalid position given in <cell> tag. Please specify a number from 1 to 9.");
					return false;
				}
				--pos;
				QString name = attributes.value("name");
				if (name.isEmpty()) {
					setError("No name specified in <cell> tag.");
					return false;
				}
				if (m_menu->cell(pos)) {
					setError("Cannot add cell \"" + name + "\" to already filled cell position " + QString::number(pos) + ".");
					return false;
				}
				m_cell = new CellData(name);
				m_menu->setCell(pos,m_cell);
				m_state = StateCell;
				return true;
			}
			else {
				setError("Only <cell> tags are allowed within a <menu> or <rootmenu> tag. Found <" + qName + "> tag instead.");
				return false;
			}
		//anything else is invalid.
		case StateExec:
			setError("Only text is allowed within an <exec> tag. Found <" + qName + "> tag instead.");
			return false;
		case StateDone:
			setError("<" + qName + "> tag found after </rootmenu> tag");
			return false;
		default:
			setError("Internal state error.");
			return false;
	}
}

bool FileParser::endElement(const QString &, const QString &,
							const QString &qName) {
	switch (m_state) {
		case StateNone:
			setError("Menu file must begin with a <rootmenu> tag");
			return false;
		case StateCell:
			//if we were parsing a cell and now we see a close cell tag,
			//go back to parsing a menu.
			if (qName == "cell") {
				m_state = StateMenu;
				return true;
			}
			else {
				setError("</" + qName + "> found where </cell> expected.");
				return false;
			}
		case StateMenu:
			//if we were parsing a menu and saw a close menu tag, go back to
			//parsing a cell.
			if (qName == "menu") {
				if (m_menu->parent()) {
					//newly created menus must be filled. This makes sure that
					//all nine cell positions are filled and makes drag and drop
					//MUCH simpler. I do not require every cell to be filled,
					//though any menu output by Menoku WILL be that way.
				    m_menu->fill();
					//move up to the previous menu
					m_menu = m_menu->parent();
    				m_state = StateCell;
                    return true;
				}
				else {
					setError("</menu> found where </rootmenu> expected.");
					return false;
				}
			}
			//if we closed a rootmenu tag
			else if (qName == "rootmenu") {
				//if this menu has a parent, it isn't the root!
			    if (m_menu->parent()) {
			        setError("</rootmenu> found where </menu> expected.");
			    }
				//otherwise, we've finished the file!
			    else {
			        m_menu->fill();
			        m_state = StateDone;
			        return true;
			    }
			}
			else {
				setError("</" + qName + "> found where </" +
				         ((m_menu == m_root)?"rootmenu":"menu") +
				         "> expected.");
				return false;
			}
		case StateExec:
			if (qName == "exec") {
				m_state = StateCell;
				return true;
			}
			else {
				setError("</" + qName + "> found where </exec> expected.");
				return false;
			}
		case StateDone:
			setError("<" + qName + "> tag found after </rootmenu> tag");
			return false;	
		default:
			setError("Internal state error.");
			return false;		
	}
}

bool FileParser::characters(const QString &str) {
	//ignore whitespace. Menoku prints nicely tabbed menu files and that
	//shouldn't throw off the parser.
	if (str.trimmed().isEmpty()) {
		return true;
	}
	//you only see bare text between exec tags.
	if (m_state == StateExec) {
		m_item->setExec(str);
		return true;
	}
	else {
		setError("Raw text is only allowed inside <exec> tags.");
		return false;
	}
	return true;
}

void FileParser::setDocumentLocator(QXmlLocator *locator) {
	m_locator = locator;
}

bool FileParser::error(const QXmlParseException &exception) {
	//this procedure will be called whenever I set my own errors, but it could be
	//called by the superclass, too. If the superclass gave me a parseexception,
	//use the message associated with it. If it's an error I raised, keep my
	//description of what went wrong.
    if (exception.message() != "error triggered by consumer")
        setError(exception.message());
    return false;
}

bool FileParser::fatalError(const QXmlParseException &exception) {
	//like above.
    if (exception.message() != "error triggered by consumer")
        setError(exception.message());
    return false;
}

void FileParser::setError(QString error) {
	//generate a nice error complete with line number!
	error = QString("Parser error on line ") + QString::number(m_locator->lineNumber()) + " at character " + QString::number(m_locator->columnNumber()) + ":\n" + error;
	m_error = error;
}
