#include "IconFactory.h"

#include <QSettings>
#include <QDir>
#include "DefaultSettings.h"

#include "ItemData.h"
#include "CellData.h"
#include "MenuData.h"

#include <QtDebug>

//declared in main.cpp
extern MenuData* root;

using namespace Qt;

IconFactory::IconFactory() {
    m_cache_all_icons = false;
    m_resize_all_icons = false;
    m_resize_width = 0;
    m_resize_height = 0;
    m_generic_exec_icon = "";
    m_generic_menu_icon = "";
}

QPixmap IconFactory::iconFor(ItemData* item) {
    QString icon_file_name = item->icon();
	//if we are using a cache and this icon is already loaded in the cache,
    if (m_cache_all_icons && m_icon_cache.contains(icon_file_name)) {
		//use the cached icon
        return m_icon_cache.value(icon_file_name);
    }

    QPixmap icon;

	//otherwise, try to load the file from the filesystem.
    if (icon.load(icon_file_name)) {
		//if the icon succeeded in loading and we're supposed to resize all
		//icons, resize THIS icon.
		if (m_resize_all_icons) {
	        icon = icon.scaled(m_resize_width,m_resize_height,IgnoreAspectRatio,SmoothTransformation);
		}
		//if we loaded a new icon, and we're caching icons, cache it!
        if (m_cache_all_icons) {
            m_icon_cache.insert(icon_file_name,icon);
        }
    }
	//if the above process didn't get us an icon...
    else {
		//then load the appropriate generic icon instead. These pixmaps are
		//already resized.
        if (item->type() == ItemData::ItemExec) {
            icon = m_generic_exec;
        }
        else {
            icon = m_generic_menu;
        }
    }

    return icon;
}

bool IconFactory::updateSettings() {
	//remember all the old settings so we can avoid unnecessary work.
    bool old_cache_all_icons = m_cache_all_icons;
    bool old_resize_all_icons = m_resize_all_icons;
    int old_resize_width = m_resize_width;
    int old_resize_height = m_resize_height;
    QString old_generic_exec_icon = m_generic_exec_icon;
    QString old_generic_menu_icon = m_generic_menu_icon;

	//get the new settings.
    QSettings settings;
    settings.beginGroup(ICON_OPTIONS);
    m_cache_all_icons = settings.value(CACHE_ALL_ICONS,CACHE_ALL_ICONS_DEFAULT).toBool();
    m_resize_all_icons = settings.value(RESIZE_ALL_ICONS,RESIZE_ALL_ICONS_DEFAULT).toBool();
    m_resize_width = settings.value(RESIZE_WIDTH,RESIZE_WIDTH).toInt();
    m_resize_height = settings.value(RESIZE_HEIGHT,RESIZE_HEIGHT_DEFAULT).toInt();
    m_generic_exec_icon = settings.value(GENERIC_EXEC_ICON,GENERIC_EXEC_ICON_DEFAULT).toString();
    m_generic_menu_icon = settings.value(GENERIC_MENU_ICON,GENERIC_MENU_ICON_DEFAULT).toString();

	//if we changed our minds about resizing the icons, or if we want to resize
	//icons and we changed the size they are resized to, then we need to rebuild
	//the icon cache and redraw everything.
    bool invalid = (old_resize_all_icons != m_resize_all_icons) ||
                   (m_resize_all_icons &&
                     (old_resize_width != m_resize_width) ||
                     (old_resize_height != m_resize_height));

	//if we're caching icons now,
    if (m_cache_all_icons) {
		//rebuild the cache if the icons have changed size or it's empty.
        if (invalid || !old_cache_all_icons) {
            rebuildCache();
        }
    }
    else {
		//otherwise, we don't want a cache of icons taking up memory! Delete them.
        m_icon_cache.clear();
    }

	//the cells will also need to be redrawn if the default icons changed.
	//we could be clever and only redraw the generic icons, but for now I
	//just redraw everything from scratch.
    invalid = invalid || (old_generic_exec_icon != m_generic_exec_icon) ||
                         (old_generic_menu_icon != m_generic_menu_icon);

	//if the old icons have changed, that means the generic icons are no longer
	//okay! Update them.
    if (invalid) {
		//try to load the user-specified generic icon file.
        if (!m_generic_exec.load(m_generic_exec_icon)) {
			//if that fails, use the built in fall-back icon.
			QPixmap pm(":exec.png");
            m_generic_exec = pm;
        }
		//do the same with the menu pixmap.
        if (!m_generic_menu.load(m_generic_menu_icon)) {
			QPixmap pm(":menu.png");
            m_generic_menu = pm;
        }
		//if they need to be resized, resize them.
        if (m_resize_all_icons) {
            m_generic_exec = m_generic_exec.scaled(m_resize_width,m_resize_height,IgnoreAspectRatio,SmoothTransformation);
            m_generic_menu = m_generic_menu.scaled(m_resize_width,m_resize_height,IgnoreAspectRatio,SmoothTransformation);
        }
    }
	//return true iff the cells need to be repainted.
    return invalid;
}

void IconFactory::cacheIcon(ItemData* item) {
    QString icon_file_name = item->icon();
	//don't try to load an icon that's already cached.
    if (!m_icon_cache.contains(icon_file_name)) {
        QPixmap icon;
		//try to load the specified file.
        if (icon.load(icon_file_name)) {
			//if successful, resize if neccessary
            if (m_resize_all_icons) {
                icon = icon.scaled(m_resize_width,m_resize_height,IgnoreAspectRatio,SmoothTransformation);
            }
			//and add this file to the cache.
            m_icon_cache.insert(icon_file_name,icon);
        }
    }
}

void IconFactory::clearCache() {
    m_icon_cache.clear();
}

void IconFactory::rebuildCache() {
    m_icon_cache.clear();
    rebuildMenu(root);
}

void IconFactory::rebuildMenu(MenuData* menu) {
	//for every cell in the menu
    for (int i = 0; i < 9; i++) {
        CellData* cell = menu->cell(i);
		//look at every location in that cell.
        for (int i = 0; i < 9; i++) {
            ItemData* item = cell->item(i);
			//if there's an item in that spot
            if (item) {
				//try to cache that icon.
                cacheIcon(item);
				//if this item contains a menu, recursively cache the icons
				//in THAT menu, too.
                if (item->type() == ItemData::ItemMenu) {
                    rebuildMenu(item->menu());
                }
            }
        }
    }
}

