#include "MenuWidget.h"
#include "MenuData.h"
#include "CellWidget.h"
#include "Arbitrator.h"

#include <QLayout>
#include <QLabel>
#include <QKeyEvent>
#include <QMessageBox>

using namespace Qt;

MenuWidget::MenuWidget(QWidget* parent)
 : QWidget(parent) {
    m_selected = false;

    QGridLayout* layout = new QGridLayout(this);
    layout->setMargin(0);
    layout->setSpacing(0);

	//set up the path label and give it a larger font to write with.
    m_path_label = new QLabel();
	QFont font;
	font.setPointSize(font.pointSize() + 4);
	m_path_label->setFont(font);
    m_path_label->setAlignment(AlignCenter);
    layout->addWidget(m_path_label,0,0,1,3);

	//instantiate all the cells and make some important connections. This lets
	//the cells and the menu seemlessly pass control from one another and
	//act as one.
    for (int i = 0; i < 9; i++) {
        m_cells[i] = new CellWidget;
        connect(m_cells[i], SIGNAL(loadMenu(MenuData*)),
                this,         SLOT(loadMenu(MenuData*)));
        connect(m_cells[i], SIGNAL(up()),
                this,         SLOT(up()));
        connect(m_cells[i], SIGNAL(select(CellWidget*)),
                this,         SLOT(select(CellWidget*)));
        connect(m_cells[i], SIGNAL(deselect()),
                this,         SLOT(deselect()));
        m_cells[i]->setEdit(false);
        //the 3-(i/3) is to convert from 2..0 to 1..3 and therefore put the
		//cells into the right spot.
        layout->addWidget(m_cells[i],3-(i/3), i%3);
    }
	//pass on all of these cells to the Arbitrator
    boss->supplyCells(m_cells);

	//Give the rows of cells reign of the space so they can push the pathlabel
	//off to the very top of the widget.
    layout->setRowStretch(1,1);
    layout->setRowStretch(2,1);
    layout->setRowStretch(3,1);
}

void MenuWidget::loadMenu(MenuData* menu) {
	//load the menu given. DO NOT DELETE ANYTHING. We need the old data to still
	//be around in case the user wants to back up or something.
    m_data = menu;
    for (int i = 0; i < 9; i++) {
        m_cells[i]->loadCell(m_data->cell(i));
    }
	//update the path display
    m_path_label->setText(m_data->path());
	//and let the Arbitrator know what happend.
    boss->menuChanged();
}

void MenuWidget::up() {
	//load the parent of the current menu (unless we're at the root)
    if (m_data->parent()) {
        loadMenu(m_data->parent());
    }
}

void MenuWidget::select(CellWidget* cell) {
	//go through every cell in the menu.
    for (int i = 0; i < 9; i++) {
		//if that's the cell we want, give it keyboard focus
        if (m_cells[i] == cell) {
            m_cells[i]->setFocus();
        }
		//otherwise, make it dark so the user knows which cell he's working with.
        else {
            m_cells[i]->setDark(true);
        }
    }
	//remember that we have something chosen.
    m_selected = true;
}

void MenuWidget::deselect() {
	//clear everything.
    for (int i = 0; i < 9; i++) {
        m_cells[i]->setDark(false);
    }
	//take back the keyboard focus
    setFocus();
    m_selected = false;
}

void MenuWidget::keyPressEvent( QKeyEvent * e )
{
	//much like the same procedure in CellWidget
	int index = 0;
	switch (e->key()) {
        case Key_Escape:
            emit( toggleVisible() );
            return;
		case Key_0:
		case Key_Insert:
			//0 either cancels a selection (undarkens all the cells) or goes
			//up one menu, depending on which is more appropriate.
			if (m_selected) {
			    deselect();
			}
			else {
			    up();
			}
			return;
		//otherwise, choose one of the cells.
		case Key_1:
		case Key_End:
			index = 0;
			break;
		case Key_2:
		case Key_Down:
			index = 1;
			break;
		case Key_3:
		case Key_PageDown:
			index = 2;
			break;
		case Key_4:
		case Key_Left:
			index = 3;
			break;
		case Key_5:
		case Key_Clear:
			index = 4;
			break;
		case Key_6:
		case Key_Right:
			index = 5;
			break;
		case Key_7:
		case Key_Home:
			index = 6;
			break;
		case Key_8:
		case Key_Up:
			index = 7;
			break;
		case Key_9:
		case Key_PageUp:
			index = 8;
			break;
		default:
            return;
	}
	
	m_cells[index]->choose();
}
