#include "SettingsWidget.h"
#include "ItemData.h"
#include "CellData.h"
#include "IconFactory.h"
#include "Arbitrator.h"
#include "MenuData.h"
#include "DefaultSettings.h"
#include <QMessageBox>
#include <QSettings>

SettingsWidget::SettingsWidget(QWidget* parent)
    : QWidget(parent) {
	
	//construct the widget!
    setupUi(this);
	//set the version on the about page.
	versionLabel->setText("Version "APP_VERSION);
    connect(applyButton,SIGNAL(clicked()),
            this,SLOT(applyClick()));
    connect(cancelButton,SIGNAL(clicked()),
            this,SLOT(cancelClick()));
    connect(deleteButton,SIGNAL(clicked()),
            this,SLOT(deleteClick()));
	//Set up the clipboard cell
    m_cell = new CellData("Clipboard");
    clipboardCell->setEdit(true);
    clipboardCell->loadCell(m_cell);
}

SettingsWidget::~SettingsWidget() {
	//since the clipboard is separate from the main menu system, it will not
	//be deleted by deleting the root menu.
    delete m_cell;
}

void SettingsWidget::setEdit(bool edit) {
    if (edit) {
		//make sure nothing's selected.
        clearItem();
		//load all the settings from QSettings!
        QSettings settings;
        settings.beginGroup(CELL_OPTIONS);
        quickSelectBox->setChecked(
            settings.value(    SPEED_SINGLE_ITEM,  SPEED_SINGLE_ITEM_DEFAULT    ).toBool());
        zoomBox->setChecked(
            settings.value(    ZOOM_SINGLE_ITEM,   ZOOM_SINGLE_ITEM_DEFAULT     ).toBool());
        hideBox->setChecked(
            settings.value(    HIDE_EMPTY_CELL,    HIDE_EMPTY_CELL_DEFAULT      ).toBool());
        settings.endGroup();

        settings.beginGroup(ICON_OPTIONS);
        cacheBox->setChecked(
            settings.value(    CACHE_ALL_ICONS,    CACHE_ALL_ICONS_DEFAULT      ).toBool());
        resizeBox->setChecked(
            settings.value(    RESIZE_ALL_ICONS,   RESIZE_ALL_ICONS_DEFAULT     ).toBool());
        widthSpin->setValue(
            settings.value(    RESIZE_WIDTH,       RESIZE_WIDTH_DEFAULT         ).toInt());
        heightSpin->setValue(
            settings.value(    RESIZE_HEIGHT,      RESIZE_HEIGHT_DEFAULT        ).toInt());
        execIconEdit->setText(
            settings.value(    GENERIC_EXEC_ICON,  GENERIC_EXEC_ICON_DEFAULT    ).toString());
        menuIconEdit->setText(
            settings.value(    GENERIC_MENU_ICON,  GENERIC_MENU_ICON_DEFAULT    ).toString());
        settings.endGroup();
		
		settings.beginGroup(WINDOW_OPTIONS);
		shortcutEdit->setText(
			settings.value(    HOT_KEY,            HOT_KEY_DEFAULT              ).toString());
		settings.endGroup();
    }
    else {
		//when we close, the side panel, save all settings.
        QSettings settings;
        settings.beginGroup(CELL_OPTIONS);
        settings.setValue(SPEED_SINGLE_ITEM,quickSelectBox->isChecked());
        settings.setValue(ZOOM_SINGLE_ITEM,zoomBox->isChecked());
        settings.setValue(HIDE_EMPTY_CELL,hideBox->isChecked());
        settings.endGroup();

        settings.beginGroup(ICON_OPTIONS);
        settings.setValue(CACHE_ALL_ICONS,cacheBox->isChecked());
        settings.setValue(RESIZE_ALL_ICONS,resizeBox->isChecked());
        settings.setValue(RESIZE_WIDTH,widthSpin->value());
        settings.setValue(RESIZE_HEIGHT,heightSpin->value());
        settings.setValue(GENERIC_EXEC_ICON,execIconEdit->text());
        settings.setValue(GENERIC_MENU_ICON,menuIconEdit->text());
        settings.endGroup();
		
		//if the hotkey has changed, try to reset it.
		settings.beginGroup(WINDOW_OPTIONS);
		QString oldKey = settings.value(HOT_KEY,HOT_KEY_DEFAULT).toString();
		QString newKey = shortcutEdit->text();
		if (oldKey != newKey) {
			settings.setValue(HOT_KEY,newKey);
			emit(updateHotKey(newKey));
		}
		settings.endGroup();
    }
}

void SettingsWidget::resizeEvent(QResizeEvent*) {
    //I'm pretty sure this should happen automatically, but since it doesn't seem
	//to...
    clipboardCell->setFixedHeight(clipboardCell->heightForWidth(width() - 2*toolBox->widget(0)->layout()->margin()));
}

void SettingsWidget::setEditEnabled(bool enabled) {
	//I should instead go back and make a frame to hold all of the edit widgets
	//and disable/enable that frame, but that's kind of a pain to do with
	//Qt Designer. So, for the time being I just go through all the Widgets
	//on the first page and disable them (except for the clipboard cell)
    QList<QObject*> children = toolBox->widget(0)->children();
    for (int i = 0; i < children.size(); ++i) {
        if (children.at(i)->isWidgetType() && children.at(i) != clipboardCell) {
            ((QWidget*)children.at(i))->setEnabled(enabled);
        }
    }
}

void SettingsWidget::setItem(ItemData* item) {
	//update all the widgets to display the data in this ItemData object
    m_item = item;
    nameEdit->setText(m_item->name());
    iconEdit->setText(m_item->icon());
    if (m_item->type() == ItemData::ItemExec) {
        execRadio->setChecked(true);
        execEdit->setText(m_item->exec());
    }
    else {
        menuRadio->setChecked(true);
        execEdit->setText("");
    }
    toolBox->setCurrentIndex(0);
	//and allow editing.
    setEditEnabled(true);
}

void SettingsWidget::clearItem() {
	//when a selection is canceled, clear all of these widgets since they're no
	//longer tied to any ItemData
    nameEdit->clear();
    iconEdit->clear();
    execRadio->setChecked(true);
    execEdit->clear();
    toolBox->setCurrentIndex(0);
	//and don't let the user edit them.
    setEditEnabled(false);
}

void SettingsWidget::applyClick() {
    if (execRadio->isChecked()) {
		//if they're trying to change a menu into an application launcher,
        if (m_item->type() == ItemData::ItemMenu) {
			//warn them that they'll be losing a lot of data.
            if (QMessageBox::question(0,
                                      "Apply changes?",
                                      "Applying changes now would cause the menu that\nused to fill this position to be lost along with\nall of its contents. Do you want to do this?",
                                      QMessageBox::Yes,QMessageBox::No)
                == QMessageBox::No) return;
        }
		//if they said okay, go ahead and delete it.
        m_item->setExec(execEdit->text());
    }
    else {
		//if this was already a menu,
        if (m_item->type() == ItemData::ItemMenu) {
			//it needs to be renamed (the menu knows its own name, even though
			//it's just inherited from the ItemData that opens the menu
            m_item->menu()->setName(nameEdit->text());
        }
        else {
			//if this was an application, we need to build a new menu!
            MenuData* menu = new MenuData(nameEdit->text(),boss->menu());
			//fill it up with empty cells (this is important! We must have nine
			//cells in every menu!)
            menu->fill();
            m_item->setMenu(menu);
        }
    }
    m_item->setName(nameEdit->text());
    m_item->setIcon(iconEdit->text());
	//let the arbitrator make sure that this item gets where it's supposed to be.
    boss->applyItem();
}

void SettingsWidget::cancelClick() {
    boss->cancelItem();
}

void SettingsWidget::deleteClick() {
	//warn if the user is trying to delete a menu, since he might be losing a lot
	//of data.
    if ((m_item->type() == ItemData::ItemMenu) &&
        QMessageBox::question(0,
                              "Delete item?",
                              "Deleting this item would cause the menu it refered\nto to be lost along with all of its contents.\nDo you want to do this?",
                              QMessageBox::Yes,QMessageBox::No)
        == QMessageBox::No) return;
    boss->deleteItem();
}

