
#include <QApplication>
#include <QMessageBox>
#include <QDesktopWidget>
#include <QSettings>
#include <QDir>
#include <iostream>
#include "DefaultSettings.h"
#include "FileParser.h"
#include "MenokuWidget.h"
#include "Arbitrator.h"
#include "IconFactory.h"
#include "MenokuTray.h"
#include "MenuData.h"

//for cout when delivering messages about command-line arguments.
using namespace std;

//the globally shared Arbitrator instance
Arbitrator* boss;
//the globally shared icon factory (needs to go here because QPixmaps cannot
//be created before there is an instance of QApplication)
IconFactory* iconFactory;

//the root of the menu being displayed
MenuData* root;
//the file where that menu is saved
QString menuFile;

int main( int argc, char **argv) {
	QApplication app(argc, argv);
	
	bool show = false;
	//technically, the Qt4.0 version of this code will work in 4.1, but they made
	//such a big deal about doing it the "proper" way in the Qt4.1 docs I decided
	//to do it their way. The only reason I have the Qt4.0 version still here is
	//that Qt4.1 doesn't quite get along with my current Linux setup.
	#if QT_VERSION >= 0x040100
		QStringListIterator it(app.arguments());
		QString called = it.next();
		//go through all the arguments
		while (it.hasNext()) {
			QString arg = it.next();
			int result = 0;
			//if the user specified a menu file to use,
			if (arg == "--menu") {
				//use that menu.
				if (it.hasNext()) {
					menuFile = it.next();
					continue;
				}
				else {
					cout << "No file name specified after --menu option." << endl;
					result = 1;
				}
			}
			//if the user wants the window to popup right away, do that.
			else if (arg == "--popup") {
				show = true;
				continue;
			}
			//if they gave any other arguments, complain.
			else if (arg != "--help") {
				cout << "Unrecognized argument: " << arg.toStdString() << endl;
				result = 1;
			}
			//if they gave a bad argument or asked for help, give them the help
			//message:
			cout << endl << "Usage: " << called.toStdString()
		    	 << " [--menu <filename> ] [--popup] [--help]" << endl
            	 << endl
            	 << "Menoku Menu System (Version "APP_VERSION")" << endl
            	 << endl
            	 << "Options:" << endl
            	 << "  --menu <filename>  Chooses the filename of the menu to show." << endl
            	 << "  --popup            Forces the Menoku window to appear on startup." << endl
            	 << "  --help             Shows this message." << endl;
    		return result;
		}		
	#else
		//this is the same as above, just accessing the arguments slightly differently
		QString called = app.argv()[0];
		for (int i = 1; i < app.argc(); i++) {
			QString arg = app.argv()[i];
			int result = 0;
			if (arg == "--menu") {
				if (++i < app.argc()) {
					menuFile = app.argv()[i];
					continue;
				}
				else {
					cout << "No file name specified after --menu option." << endl;
					result = 1;
				}
			}
			else if (arg == "--popup") {
				show = true;
				continue;
			}
			else if (arg != "--help") {
				cout << "Unrecognized argument: " << arg.toStdString() << endl;
				result = 1;
			}
			cout << endl << "Usage: " << called.toStdString()
		    	 << " [--menu <filename> ] [--popup] [--help]" << endl
            	 << endl
            	 << "Menoku Menu System (Version "APP_VERSION")" << endl
            	 << endl
            	 << "Options:" << endl
            	 << "  --menu <filename>  Chooses the filename of the menu to show." << endl
            	 << "  --popup            Forces the Menoku window to appear on startup." << endl
            	 << "  --help             Shows this message." << endl;
    		return result;
		}
	#endif
	
	//if the user didn't specify a menu on the command line,
	if (menuFile.isEmpty()) {
		//try to load the default menu.
		QDir dir = QDir::home();
		//try to pick an appropriate directory depending on OS:
		#ifdef Q_WS_WIN
		#define MENOKU_DIR "Menoku"
		#else
		#define MENOKU_DIR ".menoku"
		#endif
		//if that directory isn't already there, try to make it.
		if (!dir.exists(MENOKU_DIR)) {
			dir.mkdir(MENOKU_DIR);
		}
		if (dir.cd(MENOKU_DIR)) {
			//if the default menu isn't already there,
			menuFile = dir.filePath("menu.xml");
			//try putting an empty menu there. This means that the user will
			//get an empty menu the first time he starts Menoku.
			if (!QFile(menuFile).exists()) {
				QFile(":min.xml").copy(menuFile);
				QFile(menuFile).setPermissions(QFile::ReadOwner | QFile::WriteOwner);
			}
		}
		//if any of the above failed,
		if (!QFile(menuFile).exists()) {
			//then we have no filename to go from! No way to load a menu.
			cout << "No menu file specified and couldn't find the default menu." << endl
			     << "The default menu is: "
				 << QDir(QDir::home().filePath(MENOKU_DIR)).filePath("menu.xml").toStdString() << endl;
			return 1;
		}
	}
	
	//Try to read the menu file.
	FileParser parser;
	root = parser.parse(menuFile);
    if (!root) {
		//if there was any error, report it. At this point I'm switching to
		//GUI errors instead of console errors. Maybe I should make the switch
		//sooner? Later?
        QMessageBox::critical(0,
                              "Error loading menu file (" + menuFile + "):",
                              parser.errorString(),
                              QMessageBox::Ok,QMessageBox::NoButton);
        return 1;
    }

	//Register application info for QSettings.
	//This program is no way associated with Brown University except that the
	//author is getting his degree there  ;)
    QCoreApplication::setOrganizationName("Brown University");
    QCoreApplication::setApplicationName("Menoku");

    //now that the menu is loaded, we can preload all the icons if that's
	//what the user wants.
    iconFactory = new IconFactory();
    iconFactory->updateSettings();

	//build the arbitrator.
	boss = new Arbitrator();

	//build the main window
	MenokuWidget menoku(root);

	//load geometry settings, if they're available.
	QSettings settings;
	settings.beginGroup(WINDOW_OPTIONS);
	QSize size = settings.value(WINDOW_SIZE,WINDOW_SIZE_DEFAULT).toSize();
	QPoint pos = settings.value(WINDOW_POS,WINDOW_POS_DEFAULT).toPoint();
	//if there is no size specified, pick a decent size to fill the screen
	//but have good proportions and leave a little space for menus and panels
	//and the like. Not very scientific, but very effective.
    if (size.isEmpty()) {
        QRect avail = QDesktopWidget().availableGeometry();
		int side = min(((avail.width() - 100) / 4) * 3,avail.height() - 100);
        size.setHeight(side);
        size.setWidth(side);
    }
	//move the main window and resize it.
	menoku.resize(size);
	menoku.move(pos);
	//if the user asked for it, show the window. Otherwise, just let it sit
	//in the system tray until the user requests it.
	if (show) menoku.show();

	//create the system tray icon.
    MenokuTray tray;
    tray.show();
	boss->supplyTray(&tray);

	//and start the program!
	int result = app.exec();

	//by now the user has quit. If the application is in edit mode, switch
	//back out of it! This makes sure the latest changes are saved and that
	//the geometry that gets saved is accurate.
    if (boss->edit()) {
        boss->toggleEdit();
    }
	//save the window geometry
	settings.setValue(WINDOW_SIZE,menoku.size());
	settings.setValue(WINDOW_POS,menoku.pos());

	//clean up all the objects I made...
	delete root;
	delete iconFactory;
	delete boss;
	
	//and quit.
	return result;
}
